@@ -37,23 +37,31 @@ package URI::gemini {
 }
 
 package Net::Gemini;
-our $VERSION = '0.06';
+our $VERSION = '0.07';
 use strict;
 use warnings;
 use Encode ();
+use Exporter 'import';
 use IO::Socket::SSL;
 use Net::SSLeay;
+use Parse::MIME 'parse_mime_type';
 
-sub _DEFAULT_BUFSIZE () { 4096 }
+our @EXPORT_OK = qw(gemini_request);
 
-sub code    { $_[0]{_code} }      # 0..6 response code
-sub content { $_[0]{_buf} }
-sub error   { $_[0]{_error} }     # error message for 0 code
+sub _DEFAULT_BUFSIZE ()        { 4096 }
+sub _DEFAULT_MAX_CONTENT ()    { 2097152 }
+sub _DEFAULT_REDIRECTS ()      { 5 }
+sub _DEFAULT_REDIRECT_SLEEP () { 1 }
+
+sub code    { $_[0]{_code} }       # 0..6 response code
+sub content { $_[0]{_content} }    # NOTE only after certain calls and codes
+sub error   { $_[0]{_error} }      # error message for 0 code
 sub host    { $_[0]{_host} }
 sub meta    { $_[0]{_meta} }
+sub mime    { $_[0]{_mime} }       # NOTE only after certain calls and codes
 sub port    { $_[0]{_port} }
 sub socket  { $_[0]{_socket} }
-sub status  { $_[0]{_status} }    # two digit '1x', '2x', ... response code
+sub status  { $_[0]{_status} }     # two digit '1x', '2x', ... response code
 sub uri     { $_[0]{_uri} }
 
 # see VERIFICATION below; the caller should supply a custom callback.
@@ -89,8 +97,8 @@ sub get {
     # gemini://alexschroeder.ch/page/2020-07-20%20Does%20a%20Gemini%20certificate%20need%20a%20Common%20Name%20matching%20the%20domain%3F
     eval {
         $obj{_socket} = IO::Socket::SSL->new(
-            SSL_hostname        => $obj{_host},    # SNI
-			( $param{tofu} ? ( SSL_verifycn_scheme => 'none' ) : () ),
+            SSL_hostname => $obj{_host},    # SNI
+            ( $param{tofu} ? ( SSL_verifycn_scheme => 'none' ) : () ),
             SSL_verify_callback => sub {
                 my ( $ok, $ctx_store, $certname, $error, $cert, $depth ) = @_;
                 if ( $depth != 0 ) {
@@ -145,6 +153,10 @@ sub get {
         $obj{_buf} .= $buf;
         last if length $obj{_buf} >= 3;
     }
+    # NOTE this is sloppy; there are fewer "full two digit status codes"
+    # defined in the appendix, e.g. only 10, 11, 20, 30, 31, 40, ...
+    # on the other hand, this supports any new extensions to the
+    # existing numbers
     if ( $obj{_buf} =~ m/^(([1-6])[0-9])[ ]/ ) {
         @obj{qw(_status _code)} = ( $1, $2 );
         substr $obj{_buf}, 0, 3, '';
@@ -165,7 +177,7 @@ sub get {
             my $len = length $obj{_meta};
             if ( $len == 0 ) {
                 # special case mentioned in the specification
-                $obj{_meta} = 'text/gemini; charset=utf-8';
+                $obj{_meta} = 'text/gemini;charset=utf-8' if $obj{_code} == 2;
             } else {
                 eval {
                     $obj{_meta} = Encode::decode( 'UTF-8', $obj{_meta}, Encode::FB_CROAK );
@@ -174,8 +186,16 @@ sub get {
                     @obj{qw(_code _error)} = ( 0, "failed to decode meta: $@" );
                     goto BLESSING;
                 };
-                substr $obj{_buf}, 0, $len + 2, '';    # +2 for the \r\n
+                # another special case (RFC 2045 says that these things
+                # are not case sensitive, hence the (?i) despite the
+                # gemini specification saying "text/")
+                if (    $obj{_code} == 2
+                    and $obj{_meta} =~ m{^(?i)text/}
+                    and $obj{_meta} !~ m/(?i)charset=/ ) {
+                    $obj{_meta} .= ';charset=utf-8';
+                }
             }
+            substr $obj{_buf}, 0, $len + 2, '';    # +2 for the \r\n
             last;
         } else {
             my $len = length $obj{_buf};
@@ -202,6 +222,84 @@ sub get {
     bless( \%obj, $class ), $obj{_code};
 }
 
+# utility function that handles redirects and various means of content
+# collection
+sub gemini_request {
+    my ( $source, %options ) = @_;
+    $options{max_redirects} = _DEFAULT_REDIRECTS
+      unless exists $options{max_redirects};
+    $options{redirect_delay} = _DEFAULT_REDIRECT_SLEEP
+      unless exists $options{redirect_delay};
+    $options{max_size} = _DEFAULT_MAX_CONTENT
+      unless exists $options{max_size};
+
+    my ( $gem, $code );
+    my $redirects = 0;
+  REQUEST:
+    ( $gem, $code ) = Net::Gemini->get( $source,
+        ( exists $options{param} ? %{ $options{param} } : () ) );
+    if ( $code == 2 ) {
+        my $len     = length $gem->{_buf};
+        my $bufsize = $options{bufsize} || _DEFAULT_BUFSIZE;
+        # this can make uninit noise for a meta of ";" which might be
+        # worth an upstream patch?
+        $gem->{_mime} = [ parse_mime_type( $gem->meta ) ];
+        if ( exists $options{content_callback} ) {
+            if ($len) {
+                $options{content_callback}->( $gem->{_buf}, $len, $gem ) or goto CLEANUP;
+            }
+            while (1) {
+                my $buf;
+                $len = sysread $gem->{_socket}, $buf, $bufsize;
+                if ( !defined $len ) {
+                    die "sysread failed: $!\n";
+                } elsif ( $len == 0 ) {
+                    last;
+                }
+                $options{content_callback}->( $buf, $len, $gem ) or goto CLEANUP;
+            }
+        } else {
+            if ($len) {
+                if ( $len > $options{max_size} ) {
+                    $gem->{_content} = substr $gem->{_buf}, 0, $options{max_size};
+                    @{$gem}{qw(_code _error)} = ( 0, 'max_size' );
+                    goto CLEANUP;
+                }
+                $gem->{_content} = $gem->{_buf};
+                $options{max_size} -= $len;
+            }
+            while (1) {
+                my $buf;
+                $len = sysread $gem->{_socket}, $buf, $bufsize;
+                if ( !defined $len ) {
+                    die "sysread failed: $!\n";
+                } elsif ( $len == 0 ) {
+                    last;
+                }
+                if ( $len > $options{max_size} ) {
+                    $gem->{_content} .= substr $buf, 0, $options{max_size};
+                    @{$gem}{qw(_code _error)} = ( 0, 'max_size' );
+                    goto CLEANUP;
+                }
+                $gem->{_content} .= $buf;
+                $options{max_size} -= $len;
+            }
+        }
+    } elsif ( $code == 3 and ++$redirects <= $options{max_redirects} ) {
+        # a '31' permanent redirect should result in us not requesting
+        # the old URL again, but that would require more code here for
+        # something that is probably rare
+        my $new = $gem->{_meta};
+        $source = URI->new_abs( $new, $gem->{_uri} );
+        select( undef, undef, undef, $options{redirect_delay} );
+        goto REQUEST;
+    }
+  CLEANUP:
+    undef $gem->{_buf};
+    close $gem->{_socket};
+    return $gem, $code;
+}
+
 # drain what remains (if anything) via a callback interface. assumes
 # that a ->get call has been made
 sub getmore {
@@ -210,9 +308,10 @@ sub getmore {
     my $len = length $self->{_buf};
     if ($len) {
         $callback->( $self->{_buf}, $len ) or return;
+        undef $self->{_buf};
     }
 
-    my $bufsize = $param{bufsize} || 4096;
+    my $bufsize = $param{bufsize} || _DEFAULT_BUFSIZE;
     while (1) {
         my $buf;
         $len = sysread $self->{_socket}, $buf, $bufsize;
@@ -242,7 +341,7 @@ Net::Gemini - a small gemini client
   match($code : ==) {
     case(0) { die "request failed " . $gem->error }
     case(1) { ... $gem->meta as prompt for input ... }
-    case(2) { ... $gem->meta and $gem->content and ... }
+    case(2) { ... $gem->meta and collect on the socket ... }
     case(3) { ... $gem->meta as redirect ... }
     case(4) { ... $gem->meta as temporary failure ... }
     case(5) { ... $gem->meta as permanent failure ... }
@@ -252,7 +351,7 @@ Net::Gemini - a small gemini client
 
 =head1 DESCRIPTION
 
-This module implements code that may help implement a gemini
+This module implements code that may help implement a gemini protocol
 client in Perl.
 
 =head2 CAVEATS
@@ -265,6 +364,76 @@ similar means of verifying the other end.
 
 L<gemini://makeworld.space/gemlog/2020-07-03-tofu-rec.gmi>
 
+=head1 FUNCTION
+
+=over 4
+
+=item B<gemini_request> I<URI> [ options ... ]
+
+A utility function that is not exported by default; it calls the B<get>
+method and handles redirects and the collection of content, if any.
+
+  use Net::Gemini 'gemini_request';
+  my ( $gem, $code ) = gemini_request( 'gemini://...', ... );
+
+A code C<2> will result in the B<mime> accessor being populated with the
+Content-Type by way of the C<parse_mime_type> function from L<Parse::MIME>.
+
+A notable difference here is that a code of C<3> indicates that too many
+redirects were encountered, not that there was a redirect. This should
+be considered an error.
+
+The socket will be closed when this call ends.
+
+Options include:
+
+=over 4
+
+=item I<bufsize> => I<strictly-positive-integer>
+
+Buffer size to use for reads from the socket. 4096 by default.
+
+=item I<content_callback> => I<code-reference>
+
+Custom callback to handle one or more portions of the request content
+with, same as the B<getmore> interface. If a callback is not provided
+the content will be collected into the object via the B<content>
+accessor. The callback is given the current buffer (raw), the length of
+that buffer, and a reference to the gemini object.
+
+  gemini_request( $uri, content_callback => sub {
+    my ( $buffer, $length, $gem ) = @_;
+    ...
+    return 1;
+  });
+
+Processing will stop if the callback returns a false value.
+
+=item B<max_size> => I<strictly-positive-integer>
+
+Maximum content size to collect into B<content>. Ignored if a custom
+callback is provided. The code will be zero and the error will be
+C<max_size> and the status will start with C<2> and the content will be
+truncated if the response is larger than permitted.
+
+=item I<max_redirects> => I<strictly-positive-integer>
+
+How many redirections should be followed. C<5> is the default.
+
+=item I<redirect_delay> => I<floating-point>
+
+How long to delay between redirects, by default C<1> second. There is a
+delay by default because gemini servers or firewalls may rate limit
+requests, or the gemini server simply may not have much CPU available.
+
+=item I<param> => I<hash-reference>
+
+Parameters that will be passed to the B<get> method.
+
+=back
+
+=back
+
 =head1 METHODS
 
 =over 4
@@ -280,8 +449,9 @@ failed--or otherwise a gemini code in the range of C<1> to C<6>
 inclusive, which will indicate the next steps any subsequent code
 should take.
 
-For code C<2> responses the response body may be split between
-B<content> and whatever remains unread in the socket, if anything.
+For code C<2> responses the response body may be split between I<_buf>
+and whatever remains unread in the socket, if anything, hence the
+B<getmore> method or the B<gemini_request> utility function.
 
 Parameters include:
 
@@ -308,6 +478,10 @@ could be used to configure e.g. the C<SSL_verify_mode> or to set a
 verification callback, or to specify a custom SNI host via
 C<SSL_hostname>.
 
+C<Timeout> can be used to set a connect timeout on the socket. However,
+a server could wedge at any point following, so it may be necessary to
+wrap a B<get> request with the C<alarm> function or similar.
+
 =item B<tofu> => I<boolean>
 
 If true, only the leaf certificate will be checked. Otherwise, the full
@@ -368,6 +542,8 @@ The I<bufsize> parameter is as for B<get>.
 
 =head1 ACCESSORS
 
+Or you can use the faster hash internals, which are not expected to change.
+
 =over 4
 
 =item B<code>
@@ -379,7 +555,8 @@ outside the gemini protocol.
 
 =item B<content>
 
-The content, if any. Raw bytes. Only if the B<code> is C<2>.
+The content, if any. Raw bytes. Only if the B<code> is C<2> and a
+suitable B<gemini_request> call has been made.
 
 =item B<error>
 
@@ -393,14 +570,20 @@ Host used for the request.
 
 Gemini meta line. Use varies depending on the code.
 
+=item B<mime>
+
+Only set by B<gemini_request> for C<2> code responses; contains an array
+reference of return values from the C<parse_mime_type> function of
+L<Parse::MIME>.
+
 =item B<port>
 
 Port used for the request.
 
 =item B<socket>
 
-Socket to the server. May not be of much use after B<getmore> is
-done with.
+Socket to the server. May not be of much use after B<getmore> is done
+with, or after B<gemini_request>.
 
 =item B<status>
 
@@ -416,13 +599,17 @@ returned from the server.
 
 =head1 BUGS
 
-None known. But it is a rather incomplete module; that may be considered
-a bug? The interface is very much subject to change.
+None known. But it is a somewhat incomplete module, and the
+specification may change, too.
 
 =head1 SEE ALSO
 
 L<gemini://gemini.circumlunar.space/docs/specification.gmi> (v0.16.1)
 
+L<gemini://gemini.thebackupbox.net/test/torture>
+
+RFC 2045
+
 =head1 COPYRIGHT AND LICENSE
 
 Copyright 2022 Jeremy Mates
