@@ -9,7 +9,7 @@
 #   trustworthy does not get caught, but the evil cannot pass through."
 
 package Net::Gemini::Server;
-our $VERSION = '0.06';
+our $VERSION = '0.07';
 use strict;
 use warnings;
 # the below code mostly cargo culted from example/ssl_server.pl
@@ -68,7 +68,8 @@ sub withforks {
                 )
             ) {
                 warn "ssl handshake failed: $SSL_ERROR\n";
-                next;
+                close $client;
+                exit;
             }
         }
         if ( $param{close_before_read} ) {
